acomb <- function(...) abind(..., along=1)

simMtx <- function(memtx){
  nn <- nrow(memtx)
  pp <- ncol(memtx)
  colnret <- rep(NA,pp+2*pp*(pp-1))
  ret <- matrix(NA, nrow = nn, ncol = 2 * pp * (pp - 1) + pp)
  ininames <- rep(NA, pp)
  if (is.null(colnames(memtx))) {
    for (i in 1:pp) {
      ininames[i] <- paste0("V", i)
    }
    colnames(memtx) <- ininames
  }
  ret[, 1:pp] <- memtx
  colnret[1:pp] <- colnames(memtx)

  cmenames <- rep(NA, 2 * pp * (pp - 1))
  count <- pp + 1
  intmtx <- matrix(NA,nrow=2*pp*(pp-1),ncol=2)
  for (i in 1:pp) {
    for (j in 1:pp) {
      if (i != j) {
        ret[, count] <- memtx[, i] * as.numeric(memtx[,
                                                      j] >= 0)
        cmenames[count - pp] <- paste0(colnames(memtx)[i],
                                       "|", colnames(memtx)[j], "+")
        intmtx[count-pp,1] <- i
        intmtx[count-pp,2] <- j
        count <- count + 1


        ret[, count] <- memtx[, i] * as.numeric(memtx[,
                                                      j] < 0)
        cmenames[count - pp] <- paste0(colnames(memtx)[i],
                                       "|", colnames(memtx)[j], "-")
        intmtx[count-pp,1] <- i
        intmtx[count-pp,2] <- j
        count <- count + 1
      }
    }
  }
  colnret[(pp + 1):ncol(ret)] <- cmenames

  colnames(ret) <- colnret
  return(list(modmtx=ret,intmtx=intmtx))
}


#Maximum lambda
lambda0.cme=function(x,y){
  max(abs(t(x)%*%(y-mean(y))))/n
}

# # Screening threshold for inner product
# scr.thresh <- function(lambda.sib, lambda.cou, gamma, tau){
#
#   nlambda.sib <- length(lambda.sib)
#   nlambda.cou <- length(lambda.cou)
#
#   #Threshold matrix
#   MM <- matrix(NA,nrow=nlambda.sib,ncol=nlambda.cou) #thresholds moving down lambda.sib
#   Sbar <- (1/n + 2 + (tau*gamma)/2)/(2/gamma*exp(-tau*gamma/2)) #maximum slope for coefficient path
#   for (i in 1:nlambda.sib){
#     for (j in 1:nlambda.cou){
#       if (i > 1){
#         betabar <- Sbar*(lambda.sib[i-1]-lambda.sib[i]) #maximum coefficient
#         if (betabar <= (min(lambda.sib[i],lambda.cou[j])*gamma) ){
#           MM[i,j] <- lambda.sib[i] * ( 1/n + exp(-tau/lambda.sib[i] * (betabar - betabar^2/(2*lambda.sib[i]*gamma)) ) * (1-betabar/(lambda.sib[i]*gamma)) ) +
#             lambda.cou[j] * exp( -tau/lambda.cou[j] * (betabar - betabar^2/(2*lambda.cou[j]*gamma)) ) * (1-betabar/(lambda.cou[j]*gamma)) -
#             (1/n)*lambda.sib[i-1]
#         }else{
#           MM[i,j] <- -2
#         }
#       }else{
#         MM[i,j] <- -1
#       }
#     }
#   }
#   #Interpolate missing thresholds
#   for (i in 2:nlambda.sib){
#     MM.rem <- MM[i,][MM[i,]>0]
#     if (length(MM.rem)==0){
#       warning("Cannot forecast thresholds in scr.thresh!")
#     }
#     if (length(MM.rem)<nlambda.cou){
#       autofit <- auto.arima(log(MM.rem))
#       MM[i,][MM[i,]<=0] <- exp(forecast(autofit,h=nlambda.cou-length(MM.rem))$mean)
#     }
#   }
#
#   return(MM)
# }

#Function for estimating CV error
cv.err <- function(xx,yy,nfolds=10,rep=10){
  n <- length(yy)
  predmat = matrix(NA, nrow=n, ncol=rep)
  # meany <- mean(yy)

  for (j in seq(rep)){
    foldid = sample(rep(seq(nfolds), length=n))
    vec <- rep(NA,n)
    for (i in seq(nfolds)){
      which = (foldid == i)
      # cv.obj <- cv.glmnet(xx[!which,],y[!which],alpha=0) #just ridge regression
      # fit.obj <- glmnet(xx[!which,],y[!which],alpha=0,lambda=0)
      # xtest <- xx[which,,drop=F]
      # vec[which] <- c(predict(fit.obj,newx=xtest,s=0))
      dat <- data.frame(y = y[!which],xx[!which,])
      glmfit <- glm(y~.,data = dat)
      xtest <- xx[which,,drop=F]
      vec[which] <- predict(glmfit,data.frame(xtest))
    }
    predmat[,j] <- vec
  }

  #Compute mean and sd of cv error
  cvmat <- abs(predmat-matrix(rep(yy,rep),nrow=n))
  cvm <- mean(cvmat)
  cvsd <- mean(apply(cvmat,1,sd))
  ret <- list(cvm=cvm,cvsd=cvsd,mat=predmat)
}

#Prediction method for CME
predict.cme <- function(cme.obj,ls.ind,lc.ind,newx){
  # xnew - no need to normalize
  ret <- array(NA, c(nrow(newx), length(ls.ind), length(lc.ind)))
  for (i in 1:length(ls.ind)){
    for (j in 1:length(lc.ind)){
      # print(paste0(i,j))
      ret[,i,j] <- cme.obj$inter[i,j] + newx%*%matrix(cme.obj$coefficients[,ls.ind[i],lc.ind[j]],ncol=1)
    }
  }
  return(ret)
}

# #Prediction method for CME
# predict.cme.mean <- function(cme.obj,ls.ind,lc.ind,xnew){
#   #xme, xcme - assume normalized
#   ret <- array(NA, c(nrow(xnew), length(lc.ind), length(ls.ind)))
#   for (i in 1:length(lc.ind)){
#     for (j in 1:length(ls.ind)){
#       # print(paste0(i,j))
#       ret[,i,j] <- mean(cme.obj$y) + xnew%*%matrix(cme.obj$coefficients[,ls.ind[j],lc.ind[i]],ncol=1)
#     }
#   }
#   return(ret)
# }

#Fitting cmenet
cmenet <- function(xme, xcme, y, gamma, tau,
                   nlambda.sib=25, nlambda.cou=25,
                   lambda.min.ratio=ifelse(length(y) < ncol(xme)+ncol(xcme), 0.01, 1e-04),
                   lambda.sib=NULL, lambda.cou=NULL, lambda.grd=T,
                   maxit = 1e3, it_reset = 1, it_warm = 3, screen.ind = TRUE){
  #lambda.grd - T if grid over lambda, F otherwise

  max.lambda <- lambda0.cme(cbind(xme,xcme),y)
  if (is.null(lambda.sib)){
    lambda.sib <- exp(seq(from = log(max.lambda), to = log(max.lambda * lambda.min.ratio), length = nlambda.sib))
    lambda.cou <- exp(seq(from = log(max.lambda), to = log(max.lambda * lambda.min.ratio), length = nlambda.cou))
  }

  xme.sc <- scale(xme,center=T)
  xme.sl <- attr(xme.sc,"scaled:scale")
  xme.ce <- attr(xme.sc,"scaled:center")
  xcme.sc <- scale(xcme,center=T)
  xcme.sl <- attr(xcme.sc,"scaled:scale")
  xcme.ce <- attr(xcme.sc,"scaled:center")

  # MM <- scr.thresh(lambda.sib,lambda.cou,gamma,tau)
  beta0 <- rep(0,ncol(xme)+ncol(xcme))
  # act_vec <- rep(1,ncol(xme)+ncol(xcme))
  ret <- cme(xme.sc, xcme.sc, y,
             lambda.sib, lambda.cou, gamma, tau,
             xme.sl, xcme.sl, beta0, max.lambda,
             maxit, it_warm, it_reset, screen.ind)


  if (lambda.grd){
    inter <- matrix(NA,nrow=length(lambda.sib),ncol=length(lambda.cou))
  }else{
    inter <- matrix(NA,nrow=length(gamma),ncol=length(tau))
  }
  xmat <- cbind(xme,xcme)
  for (a in 1:nrow(inter)){
    for (b in 1:ncol(inter)){
      inter[a,b] <- mean(y - xmat%*%ret$coefficients[,a,b])
    }
  }
  ret$inter <- inter
  return(ret)
}


#CV tuning for CME
cv.cmenet <- function(xme, xcme, y, nfolds = 10, type.measure = "mse",
                      nlambda.sib=25, nlambda.cou=20, lambda.min.ratio=ifelse(length(y) < ncol(xme)+ncol(xcme), 0.01, 1e-04), lambda = NULL,
                      ngamma=9, max.gamma=150, min.gamma=1.0001, gamma = NULL,
                      ntau = 10, min.tau = 0.001/length(y), max.tau = 0.2/length(y), tau = NULL,
                      maxit = 1e3, it_reset = 1, it_warm = 3,
                      screen.ind = (ifelse(ncol(xme)<=50,FALSE,TRUE)||ifelse(nrow(xme)<=100,FALSE,TRUE)) ){

  #center and scale model matrices
  pme <- ncol(xme)
  pcme <- ncol(xcme)
  n <- nrow(xme)
  xmat <- cbind(xme,xcme) #Full model matrix

  #Pretune using SparseNet
  nlambda <- nlambda.sib
  cv.sn <- cv.sparsenet(cbind(xme,xcme),y,nfolds=nfolds,type.measure=type.measure,
                        ngamma=ngamma, max.gamma=max.gamma, min.gamma=min.gamma,
                        nlambda=nlambda, lambda.min.ratio=lambda.min.ratio, lambda = lambda)
  sn <- sparsenet(cbind(xme,xcme),y,
                  ngamma=ngamma, max.gamma=max.gamma, min.gamma=min.gamma,
                  nlambda=nlambda, lambda.min.ratio=lambda.min.ratio, lambda = lambda)
  lambda_vec <- cv.sn$lambda
  gamma_vec <- rev(cv.sn$sparsenet.fit$gamma)
  # gamma_vec[ngamma] <- 1e3 #Otherwise it causes problems with forecasting thresholds later
  tau_vec <- exp(seq(from=log(min.tau),to=log(max.tau),length.out=ntau))
  parms.mc.sn <- cv.sn$parms.min
  parms.1se.sn <- cv.sn$parms.1se

  # Conduct cross-validation to tune weights and coupling parameters
  max.lambda <- lambda0.cme(cbind(xme,xcme),y)
  lambda.sib <- exp(seq(from = log(max.lambda), to = log(max.lambda * lambda.min.ratio), length = nlambda.sib))
  lambda.cou <- exp(seq(from = log(max.lambda), to = log(max.lambda * lambda.min.ratio), length = nlambda.cou))

  #Initialize lambdas
  cl.lsib <- lambda.sib[which.min(abs(lambda.sib - (parms.mc.sn[2])/2 ))]
  cl.lcou <- lambda.cou[which.min(abs(lambda.cou - (parms.mc.sn[2])/2 ))]
  parms1.min <- c(cl.lsib,cl.lcou)
  # parms2.min <- c(init.tau,parms.mc.sn[1])

  # ptm <- proc.time()

  ##Iterate for gamma/tau:
  predmat = array(NA, c(n, length(gamma_vec), length(tau_vec)))
  foldid = sample(rep(seq(nfolds), length = n))
  cvm <- matrix(NA,length(tau_vec),length(gamma_vec))
  coef <- simple_sparse_zero_array(dim=c(ncol(xmat),length(tau_vec),length(gamma_vec)))

  #Perform N-fold CV
  # beta0 <- rep(0,ncol(xme)+ncol(xcme))
  if (nfolds < 3){
    stop("nfolds must be bigger than 3; nfolds=10 recommended")
  }

  # predmat <- foreach (i = seq(nfolds), .packages=c("cme","forecast"),
  #                     .verbose=TRUE, .combine = "acomb", .multicombine = TRUE) %dopar%{
  for (i in seq(nfolds)){
    print(paste0("Tuning gamma and tau: fold ",i," ... "))
    which = (foldid == i)
    fitobj <- cmenet(xme=xme[!which,,drop=F],xcme=xcme[!which,,drop=F],y=y[!which],
                     lambda.sib=parms1.min[1], lambda.cou=parms1.min[1], lambda.grd=F,
                     gamma=rev(gamma_vec), tau=tau_vec,
                     maxit=maxit,it_warm=it_warm,it_reset=it_reset,screen.ind=screen.ind)
    xtest <- xmat[which,,drop=F]
    predmat[which,,] <- (predict.cme(fitobj,1:length(gamma_vec),1:length(tau_vec),xtest) - y[which])^2
  }

  #Compute mean and sd of cv error
  cvm <- apply(predmat,c(2,3),mean)
  # cvm <- t(cvm)
  whichmin = argmin(cvm) #min.cv selection
  ind2.min = whichmin
  parms2.min = c(rev(gamma_vec)[whichmin[1]], tau_vec[whichmin[2]])

  ##Iterate for lambdas:
  predmat = array(NA, c(n, nlambda.sib, nlambda.cou))
  foldid = sample(rep(seq(nfolds), length = n))
  cvm <- matrix(NA,nlambda.sib,nlambda.cou)
  coef <- simple_sparse_zero_array(dim=c(ncol(xmat),nlambda.sib,nlambda.cou))

  #Perform N-fold CV
  beta0 <- rep(0,ncol(xme)+ncol(xcme))
  if (nfolds < 3){
    stop("nfolds must be bigger than 3; nfolds=10 recommended")
  }

  for (i in seq(nfolds)){
    print(paste0("Tuning lambdas: fold ",i," ... "))
    which = (foldid == i)
    # act_vec <- rep(1,ncol(xme)+ncol(xcme))
    fitobj <- cmenet(xme=xme[!which,,drop=F], xcme=xcme[!which,,drop=F], y=y[!which],
                     lambda.sib=lambda.sib, lambda.cou=lambda.cou, gamma=parms2.min[1], tau=parms2.min[2],
                     maxit=maxit,it_warm=it_warm,it_reset=it_reset,screen.ind=screen.ind)
    xtest <- xmat[which,,drop=F]
    predmat[which,,] <- (predict.cme(fitobj,1:length(lambda.sib),1:length(lambda.cou),xtest) - y[which])^2
    # (predict.cme(fitobj,1:length(lambda.sib),1:length(lambda.cou),xtest) - y[which])^2
  }

  #Compute mean and sd of cv error
  cvm <- apply(predmat,c(2,3),mean)
  cvsd <- apply(predmat^2,c(2,3),sd)/length(y)
  # cvm <- t(cvm)
  # cvsd <- t(cvsd)
  whichmin = argmin(cvm) #min.cv selection
  ind1.min = whichmin
  parms1.min = c(lambda.sib[whichmin[1]], lambda.cou[whichmin[2]])


  #Summarize into list
  obj = list(lambda.sib = lambda.sib, lambda.cou = lambda.cou, gamma = gamma_vec, tau = tau_vec,
             err = type.measure, cvm = cvm)
  # obj$which.min = c(ind1.min[2], ind1.min[1], ind2.min[2], ind2.min[1]) #siblings, cousins, gamma, tau
  obj$parms.min = c(parms1.min[1], parms1.min[2], parms2.min[1], parms2.min[2])
  # names(obj$which.min) <- c("lambda.sib","lambda.cou","gamma","tau")
  names(obj$parms.min) <- c("lambda.sib","lambda.cou","gamma","tau")

  #1se selection
  print(paste0("Fitting full data ..."))
  fitall <- cmenet(xme=xme, xcme=xcme, y,
                   lambda.sib=lambda.sib, lambda.cou=lambda.cou, gamma=obj$parms.min[3], tau=obj$parms.min[4],
                   maxit=maxit,it_warm=it_warm,it_reset=it_reset,screen.ind=screen.ind)
  nz <- fitall$nzero
  which = cvm < min(cvm) + cvsd
  nz[!which] = 1e+40
  whichcv = argmin(nz)
  # obj$which.1se <- c(whichcv[2], whichcv[1], ind2.min[2], ind2.min[1]) #siblings, cousins, gamma, tau
  obj$parms.1se = c(lambda.sib[whichcv[1]], lambda.cou[whichcv[2]], parms2.min[1], parms2.min[2])
  obj$cme.fit <- fitall
  obj$select.min <- which(fitall$coefficients[,which(lambda.sib==obj$parms.min[1]),which(lambda.cou==obj$parms.min[2])] != 0)
  obj$select.1se <- which(fitall$coefficients[,which(lambda.sib==obj$parms.1se[1]),which(lambda.cou==obj$parms.1se[2])] != 0)
  class(obj) = "cv.cme"

  return(obj)

}
